// Env.cs
using System;
using System.Collections;


namespace TtLisp {
  public class Env {
    // -- instance -------------------------------------------------------
    private Stack         symbolStack;
    private FunctionTable functionTable;

    public Env() {
      this.symbolStack = new Stack();
      this.symbolStack.Push( new SymbolTable() );

      this.SymbolTable.ConstIntern( Nil.NIL, Nil.NIL );
      this.SymbolTable.ConstIntern( True.T, True.T );

      this.functionTable = new FunctionTable();

      this.DefineFunction( "car",             new SystemFunctions.FunctionCar() );
      this.DefineFunction( "cdr",             new SystemFunctions.FunctionCdr() );
      this.DefineFunction( "quote",           new SystemFunctions.FunctionQuote() );
      this.DefineFunction( "cons",            new SystemFunctions.FunctionCons() );
      this.DefineFunction( "atom",            new SystemFunctions.FunctionAtom() );
      this.DefineFunction( "eq",              new SystemFunctions.FunctionEq() );
      this.DefineFunction( "progn",           new SystemFunctions.FunctionProgn() );
      this.DefineFunction( "if",              new SystemFunctions.FunctionIf() );
      this.DefineFunction( "cond",            new SystemFunctions.FunctionCond() );
      this.DefineFunction( "lambda",          new SystemFunctions.FunctionLambda() );
      this.DefineFunction( "macro",           new SystemFunctions.FunctionMacro() );
      this.DefineFunction( "bind-function",   new SystemFunctions.FunctionBindFunction() );
      this.DefineFunction( "function",        new SystemFunctions.FunctionFunction() );
      this.DefineFunction( "apply",           new SystemFunctions.FunctionApply() );
      this.DefineFunction( "funcall",         new SystemFunctions.FunctionFuncall() );
      this.DefineFunction( "set",             new SystemFunctions.FunctionSet() );
      this.DefineFunction( "setq",            new SystemFunctions.FunctionSetQ() );
      this.DefineFunction( "eval",            new SystemFunctions.FunctionEval() );
      this.DefineFunction( "list",            new SystemFunctions.FunctionList() );
      this.DefineFunction( "macroexpand",     new SystemFunctions.FunctionMacroExpand() );
      this.DefineFunction( "let*",            new SystemFunctions.FunctionLetAste() );
      this.DefineFunction( "+",               new SystemFunctions.FunctionIntAdd() );
      this.DefineFunction( "-",               new SystemFunctions.FunctionIntSub() );
      this.DefineFunction( "*",               new SystemFunctions.FunctionIntMul() );
      this.DefineFunction( "/",               new SystemFunctions.FunctionIntDiv() );
      this.DefineFunction( "mod",             new SystemFunctions.FunctionIntMod() );
      this.DefineFunction( "=",               new SystemFunctions.FunctionIntEq() );
      this.DefineFunction( "/=",              new SystemFunctions.FunctionIntNe() );
      this.DefineFunction( ">",               new SystemFunctions.FunctionIntGt() );
      this.DefineFunction( "<",               new SystemFunctions.FunctionIntLt() );
      this.DefineFunction( ">=",              new SystemFunctions.FunctionIntGe() );
      this.DefineFunction( "<=",              new SystemFunctions.FunctionIntLe() );
      this.DefineFunction( "print",           new SystemFunctions.FunctionPrint() );
      this.DefineFunction( "require",         new SystemFunctions.FunctionRequire() );
      this.DefineFunction( "all-functions",   new SystemFunctions.FunctionAllFunctions() );
      this.DefineFunction( "every",           new SystemFunctions.FunctionEvery() );
      this.DefineFunction( "mapcar",          new SystemFunctions.FunctionMapCar() );
      this.DefineFunction( "and",             new SystemFunctions.FunctionAnd() );
      this.DefineFunction( "or",              new SystemFunctions.FunctionOr() );
      this.DefineFunction( "not",             new SystemFunctions.FunctionNot() );
      this.DefineFunction( "null",            new SystemFunctions.FunctionNot() );
      this.DefineFunction( "symbol-function", new SystemFunctions.FunctionSymbolFunction() );

      this.DefineFunction( "defun",
                           new Macro( List.MakeAsSymbol( "_name",
                                                         "_lambda-list",
                                                         "_body" ),
                                      List.Make( new Symbol( "list" ),
                                                 List.QuoteAsSymbol( "bind-function" ),
                                                 List.Make( new Symbol("list" ),
                                                            List.QuoteAsSymbol( "quote" ),
                                                            new Symbol( "_name" ) ),
                                                 List.Make( new Symbol( "list" ),
                                                            List.QuoteAsSymbol( "lambda" ),
                                                            new Symbol( "_lambda-list" ),
                                                            new Symbol( "_body" ) ) ) ) );

      this.DefineFunction( "defmacro",
                           new Macro( List.MakeAsSymbol( "_name",
                                                         "_lambda-list",
                                                         "_body" ),
                                      List.Make( new Symbol( "list" ),
                                                 List.QuoteAsSymbol( "bind-function" ),
                                                 List.Make( new Symbol("list" ),
                                                            List.QuoteAsSymbol( "quote" ),
                                                            new Symbol( "_name" ) ),
                                                 List.Make( new Symbol( "list" ),
                                                            List.QuoteAsSymbol( "macro" ),
                                                            new Symbol( "_lambda-list" ),
                                                            new Symbol( "_body" ) ) ) ) );

      this.DefineFunction( "unless",
                           new Macro( List.MakeAsSymbol( "_cond", "_then", "_else" ),
                                      List.Make( new Symbol( "if" ),
                                                 List.MakeAsSymbol( "not", "_cond" ),
                                                 new Symbol( "_then" ),
                                                 new Symbol( "_else" ) ) ) );
      
    }

    private void DefineFunction( string name, Function func ) {
      this.FunctionTable.Define( new Symbol( name ), func );
    }

    public SymbolicExpression Eval( SymbolicExpression s ) {
      return s.ToEvaluated( this );
    }

    private Stack SymbolStack { get { return this.symbolStack; } }
    public SymbolTable SymbolTable {
      get { return (SymbolTable)this.SymbolStack.Peek(); }
    }
    public void PushStack() {
      this.SymbolStack.Push( new SymbolTable( this.SymbolTable ) );
    }
    public void PopStack() {
      this.SymbolStack.Pop();
    }

    public FunctionTable FunctionTable {
      get { return this.functionTable; }
    }

  }
}
