// Interpreter.cs
using System;
using System.IO;

namespace TtLisp {
  public class Interpreter {
    private Env env;
    public Interpreter() {
      this.env = new Env();
    }

    public void ReadFromFile( string filename ) {
      using ( TextReader reader = new StreamReader( filename ) ) {
        this.ReadFromReader( reader );
      }
    }
    public void ReadFromReader( TextReader reader ) {
      this.ReadFromString( reader.ReadToEnd() );
    }
    public void ReadFromString( string src ) {
      foreach ( SymbolicExpression s in Parser.Parse( src ) ) {
        this.env.Eval( s );
      }
    }

    public void StartPrompt() {
      for (;;) {
        Console.Write( "Lisp> " );
        try {
          string src = Console.In.ReadLine();
          if ( String.Compare( src, "QUIT", true ) == 0 || String.Compare( src, "EXIT", true ) == 0 ) {
            break;
          }
          foreach ( SymbolicExpression s in Parser.Parse( src ) ) {
            Console.WriteLine( this.env.Eval( s ) );
          }
        }
        catch ( SymbolUnboundException e ) {
          Console.WriteLine( "ϐ`Ă܂: " + e.Symbol.Name );
        }
        catch ( FunctionIsNotDefinedException e ) {
          Console.WriteLine( "֐`Ă܂: " + e.Symbol.Name );
        }
        catch ( LispException e ) {
          Console.WriteLine( e.GetType().ToString() + "܂B" );
          // Console.WriteLine( e.StackTrace );
        }
      }
    }

    static public int Main( string[] args ) {
      Interpreter interpreter = new Interpreter();
      foreach ( string arg in args ) {
        try {
          interpreter.ReadFromFile( arg );
          Console.WriteLine( "read finished {0}", arg );
        }
        catch ( Exception e ) {
          Console.WriteLine( "in {0}", arg );
          Console.WriteLine( e.GetType().ToString() + "܂B" );
        }
      }
      interpreter.StartPrompt();
      return 0;
    }
    
  }
}
