// List.cs
using System;
using System.Text;
using System.Collections;


namespace TtLisp {
  public class List : SymbolicExpression, IEnumerable {
    // -- static ---------------------------------------------------------
    // Xg쐬ÓI\bhQ
    public static List Make( params SymbolicExpression[] args ) {
      List list = Nil.NIL;
      for ( int i = args.Length - 1; i >= 0; --i ) {
        list = new List( args[i], list );
      }
      return list;
    }

    public static List MakeAsSymbol( params string[] args ) {
      List list = Nil.NIL;
      for ( int i = args.Length - 1; i >= 0; --i ) {
        list = new List( new Symbol( args[i] ), list );
      }
      return list;
    }

    public static List Quote( SymbolicExpression s ) {
      return List.Make( new Symbol( "quote" ), s );
    }

    public static List QuoteAsSymbol( String str ) {
      return List.Quote( new Symbol( str ) );
    }

    // -- instance -------------------------------------------------------
    private readonly SymbolicExpression car;
    private readonly SymbolicExpression cdr;

    public List( SymbolicExpression car ) : this( car, Nil.NIL ) {}
    public List( SymbolicExpression car, SymbolicExpression cdr ) {
      this.car = car;
      this.cdr = cdr;
    }

    public SymbolicExpression Car {
      get { return this.car; }
    }

    public SymbolicExpression Cdr {
      get { return this.cdr; }
    }

    public SymbolicExpression this[int index] {
      get {
        List cur = this;
        for ( int i = 0; i < index; ++i ) {
          cur = (List)cur.Cdr;
        }
        return cur.Car;
      }
    }

    public int Size {
      get {
        if ( this is Nil ) { return 0; }
        if ( !(this.Cdr is List) ) { return 1; }
        return ((List)this.Cdr).Size + 1;
      }
    }

    public override SymbolicExpression ToEvaluated( Env env ) {
      Function func;
      if ( this.Car is ISymbol ) {
        func = env.FunctionTable[(ISymbol)this.Car];
      }
      else {
        SymbolicExpression _tmp = env.Eval( this.Car );
        if ( _tmp is Function ) {
          func = (Function)_tmp;
        }
        else {
          throw new InvalidFunctionException();
        }
      }
      return func.Call( env, this.Cdr is List ? (List)this.Cdr : Nil.NIL );
    }

    public override string ToString() {
      if ( this.Size == 2 && this.Car is ISymbol ) {
        switch ( ((ISymbol)this.Car).Name ) {
        case "quote":
          return String.Concat( "'", this[1].ToString() );

        case "function":
          return String.Concat( "#'", this[1].ToString() );

        default:
          break;
        }
      }

      StringBuilder sb = new StringBuilder();
      sb.Append( "(" );
      for ( List cur = this; ; cur = (List)cur.Cdr ) {
        sb.Append( cur.Car.ToString() );
        if ( cur.Cdr is Nil ) { break; }
        if ( cur.Cdr is Atom ) {
          sb.Append( " . " );
          sb.Append( cur.Cdr.ToString() );
          break;
        }
        sb.Append( " " );
      }
      sb.Append( ")" );
      return sb.ToString();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return new ListEnumeratorImple( this );
    }
    public ListEnumerator GetEnumerator() {
      return new ListEnumerator( new ListEnumeratorImple( this ) );
    }
  }

  // ---------------------------------------------------------------------
  // ȉ List  foreach ňׂ̃NX
  public class ListEnumeratorImple : IEnumerator {
    private List root;
    private List current;
    public ListEnumeratorImple( List list ) {
      this.root = new List( Nil.NIL, list );
      this.Reset();
    }
    public object Current {
      get {
        if ( this.current == this.root ) {
          try { new object[0].GetEnumerator().Current.ToString(); }
          catch ( Exception e ) { throw e; }
        }
        if ( this.current is Nil ) {
          try { IEnumerator _tmp = new object[0].GetEnumerator(); _tmp.MoveNext(); _tmp.Current.ToString(); }
          catch ( Exception e ) { throw e; }
        }
        return this.current.Car;
      }
    }

    public bool MoveNext() {
      if ( this.current.Cdr is List ) {
        this.current = (List)this.current.Cdr;
        return !(this.current is Nil);
      }
      this.current = Nil.NIL;
      return false;
    }

    public void Reset() {
      this.current = this.root;
    }
  }

  public class ListEnumerator {
    private ListEnumeratorImple enumerator;
    public ListEnumerator( ListEnumeratorImple enumerator ) {
      this.enumerator = enumerator;
    }
    public SymbolicExpression Current { get { return (SymbolicExpression)this.enumerator.Current; } }
    public bool MoveNext() { return this.enumerator.MoveNext(); }
    public void Reset() { this.enumerator.Reset(); }
  }

}
